from decimal import Decimal
from datetime import date, datetime
from typing import Optional, List, Dict, DefaultDict, Any
from py_portfolio_index.constants import CACHE_DIR
from py_portfolio_index.models import (
    RealPortfolio,
    RealPortfolioElement,
    Money,
    ProfitModel,
)
from py_portfolio_index.common import divide_into_batches
from py_portfolio_index.portfolio_providers.base_portfolio import (
    BaseProvider,
    ObjectKey,
)
from py_portfolio_index.exceptions import ConfigurationError
from py_portfolio_index.constants import Logger, UNKNOWN_TICKER
from py_portfolio_index.exceptions import OrderError
from py_portfolio_index.enums import ProviderType
from py_portfolio_index.models import DividendResult
from collections import defaultdict
from functools import lru_cache
from os import environ, remove
from pathlib import Path
from platformdirs import user_cache_dir
from pytz import UTC
from time import sleep
import re

FRACTIONAL_SLEEP = 60
BATCH_SIZE = 50
FRACTIONAL_SLEEP = 60

CACHE_PATH = "schwab_tickers.json"
CACHE_DESC_PATH = "schwab_desc_to_ticker.json"

HARD_CODED_DESC_TO_TICKER: dict[str, str] = {
    "AT&T INC": "T",
    "ALBERTSONS CO SHS CLASS CLASS A": "ACI",
    "ATLANTICA SUSTAINABLE F": "AY",
    "ATLANTICA SUSTAINABLE FMANDATORY MERGER": "AY",
    "EATON CORP PLC F": "ETN",
    "S&P GLOBAL INC": "SPGI",
    "INVESCO LTD F": "IVZ",
    "SCHWAB1 INT 10/30-11/26": "SCHW",
}


def date_to_datetimes(at_day: date) -> tuple[datetime, datetime]:
    start_datetime = datetime(
        day=at_day.day,
        year=at_day.year,
        month=at_day.month,
        hour=0,
        minute=0,
        second=0,
        microsecond=0,
        tzinfo=UTC,
    )
    end_datetime = datetime(
        day=at_day.day,
        year=at_day.year,
        month=at_day.month,
        hour=23,
        minute=59,
        second=59,
        microsecond=0,
        tzinfo=UTC,
    )
    return start_datetime, end_datetime


def api_helper(response):
    from httpx import Response

    raw: Response = response
    try:
        raw.raise_for_status()
    except Exception as e:
        if "Exception while authenticating refresh token" in str(e):
            raise ConfigurationError(str(e))
        raise e
    return raw.json()


class SchwabProvider(BaseProvider):
    """Provider for interacting with stocks held in
    Schwab
    """

    PROVIDER = ProviderType.SCHWAB
    SUPPORTS_BATCH_HISTORY = 0
    API_KEY_ENV = "SCHWAB_API_KEY"
    APP_SECRET_ENV = "SCHWAB_APP_SECRET"
    SUPPORTS_FRACTIONAL_SHARES = False
    # TRADE_TOKEN_ENV = "SCHWAB_TRADE_TOKEN"
    # DEVICE_ID_ENV = "SCHWAB_DEVICE_ID"

    def __init__(
        self,
        api_key: str | None = None,
        app_secret: str | None = None,
        skip_cache: bool = False,
        external_auth: bool = False,
    ):
        if not api_key:
            api_key = environ.get(self.API_KEY_ENV, None)
        if not app_secret:
            app_secret = environ.get(self.APP_SECRET_ENV, None)
        # if not trade_token:
        #     trade_token = environ.get(self.TRADE_TOKEN_ENV, None)
        # if not device_id:
        #     device_id = environ.get(self.DEVICE_ID_ENV, None)
        if not (api_key and app_secret):
            raise ConfigurationError(
                "Must provide ALL OF api_key and app_secret arguments or set environment variables SCHWAB_API_KEY, SCHWAB_APP_SECRET "
            )

        token_path = (
            Path(user_cache_dir(CACHE_DIR, ensure_exists=True)) / "schwab_token.json"
        )
        from schwab import auth
        from schwab.utils import Utils

        try:
            c = auth.client_from_token_file(token_path, api_key, app_secret=app_secret)
        except FileNotFoundError:
            if external_auth:
                raise ConfigurationError(
                    "External authentication flag set but invalid auth"
                )

            c = auth.client_from_login_flow(
                api_key,
                app_secret,
                "https://127.0.0.1:8182",
                token_path,
                interactive=external_auth,
            )
        # we must set both of these to have a valid login
        BaseProvider.__init__(self)
        self._provider = c
        try:
            self._account_hash = api_helper(self._provider.get_account_numbers())[0][
                "hashValue"
            ]
        except Exception as e:
            remove(token_path)
            raise ConfigurationError(
                f"Authentication is expired: {str(e)}. Removed token."
            )
        self._utils = Utils(self._provider, account_hash=self._account_hash)
        self._local_description_lookup_cache: dict[str, str] = {}
        if not skip_cache:
            self._load_local_description_lookup_cache()

    def _load_local_description_lookup_cache(self):
        from platformdirs import user_cache_dir
        from pathlib import Path
        import json

        path = Path(user_cache_dir("py_portfolio_index", ensure_exists=True))
        file = path / CACHE_DESC_PATH
        if not file.exists():
            self._local_description_lookup_cache = {}
            return
        with open(file, "r") as f:
            self._local_description_lookup_cache = json.load(f)

    def _save_local_description_lookup_cache(self):
        from platformdirs import user_cache_dir
        from pathlib import Path
        import json

        path = Path(user_cache_dir("py_portfolio_index", ensure_exists=True))
        file = path / CACHE_DESC_PATH
        with open(file, "w") as f:
            json.dump(self._local_description_lookup_cache, f)

    @lru_cache(maxsize=None)
    def _get_instrument_price(
        self, ticker: str, at_day: Optional[date] = None
    ) -> Optional[Decimal]:
        stored = self._price_cache.get_prices(tickers=[ticker], date=at_day)
        if stored:
            return stored[ticker]
        if at_day:
            start_datetime, end_datetime = date_to_datetimes(at_day)
            historicals = api_helper(
                self._provider.get_price_history_every_day(
                    symbol=ticker,
                    start_datetime=start_datetime,
                    end_datetime=end_datetime,
                )
            )
            rval = Decimal(value=historicals[0].vwap)

        else:
            quotes = api_helper(self._provider.get_quote(symbol=ticker))
            rval = Decimal(value=quotes["quotes"])
        return rval

    def _buy_instrument(
        self,
        symbol: str,
        qty: int,
        value: Optional[Money] = None,
        price: Optional[Decimal] = None,
    ) -> None:
        from schwab.orders.equities import equity_buy_market, Duration, Session
        from httpx import Request

        order: Request = self._provider.place_order(
            self._account_hash,
            order_spec=equity_buy_market(symbol, quantity=int(qty))
            .set_duration(Duration.DAY)
            .set_session(Session.NORMAL)
            .build(),
        )
        try:
            _ = self._utils.extract_order_id(order)
        except Exception as e:
            if "order not successful: status 429" in str(e):
                Logger.info(
                    f"RH error: was throttled on fractional orders! Sleeping {FRACTIONAL_SLEEP}"
                )
                sleep(FRACTIONAL_SLEEP)
                return self._buy_instrument(
                    symbol=symbol, qty=qty, value=value, price=price
                )
            raise e
        return None

    def buy_instrument(self, ticker: str, qty: Decimal, value: Optional[Money] = None):
        if qty:
            orders_kwargs: dict[str, Decimal | Money | None] = {
                "qty": qty,
                "value": None,
            }
        else:
            orders_kwargs = {"qty": None, "value": value}

        try:
            self._buy_instrument(ticker, **orders_kwargs)  # type: ignore
        except Exception as e:
            raise OrderError(f"Could not buy {ticker}: {str(e)}")
        return True

    def get_unsettled_instruments(self) -> set[str]:
        orders = []
        for status in (
            self._provider.Order.Status.PENDING_ACTIVATION,
            self._provider.Order.Status.QUEUED,
            self._provider.Order.Status.WORKING,
        ):
            orders = api_helper(
                self._provider.get_orders_for_account(
                    account_hash=self._account_hash, status=status
                )
            )
        return set(item["instrument"]["symbol"] for item in orders)

    def _get_stock_info(self, ticker: str) -> dict:
        return api_helper(
            self._provider.get_instruments(
                symbols=[ticker],
                project=self._provider.Instrument.Projection.FUNDAMENTAL,
            )
        )

    def _get_stock_info_fuzzy(self, search: str) -> dict:
        # do our best to match a description to a ticker
        STRIP_VALUES = ["FCLASS", "CLASS", "CLASS EQUITY"]
        for strip in STRIP_VALUES:
            search = search.replace(strip, "").strip()
        search = search.strip()
        search = re.sub(r"\s+", r" ", search)
        search = f"(?i){search[:20]}.*"
        search = search.replace("&", ".")
        return api_helper(
            self._provider.get_instruments(
                symbols=[search],
                projection=self._provider.Instrument.Projection.DESCRIPTION_REGEX,
            )
        )

    def get_portfolio(self) -> dict:
        from schwab.client import Client

        try:
            return api_helper(
                self._provider.get_account(
                    account_hash=self._account_hash,
                    fields=Client.Account.Fields.POSITIONS,
                )
            )["securitiesAccount"]
        except KeyError as e:
            raise ConfigurationError(
                f"Could not fetch portfolio on {str(e)}; assuming session expired"
            )
        except Exception as e:
            if "refresh_token" in str(e):
                raise ConfigurationError(
                    f"Could not fetch portfolio: {str(e)}; assuming session expired"
                )
            raise e

    def get_holdings(self) -> RealPortfolio:
        accounts_data = self._get_cached_value(
            ObjectKey.ACCOUNT, callable=self.get_portfolio
        )
        my_stocks = accounts_data["positions"]

        unsettled = self._get_cached_value(
            ObjectKey.UNSETTLED, callable=self.get_unsettled_instruments
        )

        pre = {}
        symbols = []
        for row in my_stocks:
            local: Dict[str, Any] = {}
            local["units"] = row["longQuantity"]
            instrument: dict[str, str] = row["instrument"]
            # unclear what is happening here, but skip this for now
            ticker = instrument.get("symbol", UNKNOWN_TICKER)
            local["ticker"] = ticker
            symbols.append(ticker)
            local["value"] = Money(value=row["marketValue"])
            local["weight"] = 0
            pre[ticker] = local
        prices = self._price_cache.get_prices(symbols)
        total_value = Decimal(0.0)
        for s in symbols:
            price = prices[s]
            if not price:
                continue
            total_value += price * Decimal(pre[s]["units"])
        final = []
        pl_info = self.get_per_ticker_profit_or_loss()
        for s in symbols:
            local = pre[s]
            local["weight"] = local["value"].value / total_value
            local["unsettled"] = s in unsettled
            local["appreciation"] = pl_info[s].appreciation
            local["dividends"] = pl_info[s].dividends
            final.append(local)
        out = [RealPortfolioElement(**row) for row in final]
        cash = Decimal(accounts_data["currentBalances"]["cashBalance"])
        return RealPortfolio(holdings=out, cash=Money(value=cash), provider=self)

    def _get_instrument_prices(
        self, tickers: List[str], at_day: Optional[date] = None
    ) -> Dict[str, Optional[Decimal]]:
        batches: List[Dict[str, Optional[Decimal]]] = []
        prices: Dict[str, Optional[Decimal]] = {}

        for list_batch in divide_into_batches(tickers, 100):
            if at_day:
                start_datetime, end_datetime = date_to_datetimes(at_day)
                for ticker in list_batch:
                    historicals = api_helper(
                        self._provider.get_price_history_every_day(
                            symbol=ticker,
                            start_datetime=start_datetime,
                            end_datetime=end_datetime,
                        )
                    )
                    batches.append(
                        {ticker: Decimal(value=historicals["candles"][0]["close"])}
                    )
            else:
                quotes = api_helper(self._provider.get_quotes(symbols=list_batch))
                for ticker in list_batch:
                    if ticker in quotes:
                        prices[ticker] = Decimal(
                            value=quotes[ticker]["quote"]["lastPrice"]
                        )
                    else:
                        prices[ticker] = None
        for fbatch in batches:
            prices = {**prices, **fbatch}
        return prices

    def _get_dividends_wrapper(self):
        from schwab.client.base import BaseClient

        return api_helper(
            self._provider.get_transactions(
                account_hash=self._account_hash,
                transaction_types=BaseClient.Transactions.TransactionType.DIVIDEND_OR_INTEREST,
            )
        )

    def get_per_ticker_profit_or_loss(self) -> Dict[str, ProfitModel]:
        account_info = self._get_cached_value(
            ObjectKey.ACCOUNT, callable=self.get_portfolio
        )

        dividends = self._get_dividends()

        first = {
            x["instrument"]["symbol"]: ProfitModel(
                appreciation=Money(value=Decimal(x["longOpenProfitLoss"])),
                dividends=dividends[x["instrument"].get("symbol", UNKNOWN_TICKER)],
            )
            for x in account_info["positions"]
        }
        for k, v in dividends.items():
            if k not in first:
                first[k] = ProfitModel(appreciation=Money(value=0), dividends=v)
        return first

    def _get_dividends(self) -> defaultdict[str, Money]:
        dividends: dict = self._get_cached_value(
            ObjectKey.DIVIDENDS_DETAIL, callable=self._get_dividends_wrapper
        )
        base: list[dict] = []

        changes: bool = False
        for item in dividends:
            lookup_desc = item["description"]
            ticker = self._local_description_lookup_cache.get(lookup_desc)
            if not ticker:
                ticker = HARD_CODED_DESC_TO_TICKER.get(lookup_desc)
            if not ticker:
                fuzzy_search = self._get_stock_info_fuzzy(search=lookup_desc)
                if fuzzy_search.get("instruments"):
                    match = fuzzy_search["instruments"][0]
                    self._local_description_lookup_cache[lookup_desc] = match["symbol"]
                    changes = True
                else:
                    ticker = lookup_desc
            base.append(
                {"value": Money(value=Decimal(item["netAmount"])), "ticker": ticker}
            )
        if changes:
            self._save_local_description_lookup_cache()
        final: DefaultDict[str, Money] = defaultdict(lambda: Money(value=0))
        for dividend_item in base:
            final[dividend_item["ticker"]] += dividend_item["value"]
        return final

    def get_dividend_details(
        self, start: datetime | None = None
    ) -> list[DividendResult]:
        dividends: dict = self._get_cached_value(
            ObjectKey.DIVIDENDS, callable=self._get_dividends_wrapper
        )
        final = []
        changes = False
        for item in dividends:
            lookup_desc = item["description"]
            ticker = self._local_description_lookup_cache.get(lookup_desc)
            if not ticker:
                ticker = HARD_CODED_DESC_TO_TICKER.get(
                    lookup_desc,
                )
            if not ticker:
                fuzzy_search = self._get_stock_info_fuzzy(search=lookup_desc)
                if fuzzy_search.get("instruments"):
                    match = fuzzy_search["instruments"][0]
                    self._local_description_lookup_cache[lookup_desc] = match["symbol"]
                    changes = True
                else:
                    ticker = lookup_desc
            final_ticker = ticker or UNKNOWN_TICKER
            if item["status"] == "VALID":
                event_date = datetime.fromisoformat(item["time"]).date()
                if start and event_date < start.date():
                    continue

                final.append(
                    DividendResult(
                        ticker=final_ticker,
                        amount=Money(value=float(item["netAmount"])),
                        date=event_date,
                        provider=self.PROVIDER,
                        external_id=str(item["activityId"]),
                    )
                )
        if changes:
            self._save_local_description_lookup_cache()
        return final



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from py_portfolio_index.models import LoginResponse


class PriceFetchError(Exception):
    def __init__(self, tickers: list[str] | None = None, *args):
        super().__init__(*args)
        self.tickers = tickers or []


class ConfigurationError(Exception):
    pass


class ExtraAuthenticationStepException(Exception):
    def __init__(self, response: "LoginResponse", *args):
        super().__init__(*args)
        self.response = response


class SchwabExtraAuthenticationStepException(ExtraAuthenticationStepException):
    pass


class OrderError(Exception):
    def __init__(self, message, *args):
        super().__init__(message, *args)
        self.message = message



from py_portfolio_index.models import (
    RealPortfolioElement,
    CompositePortfolio,
    Money,
    IdealPortfolio,
    IdealPortfolioElement,
)
from py_portfolio_index.enums import PurchaseStrategy
from py_portfolio_index.portfolio_providers.local_dict import (
    LocalDictProvider,
    LocalDictNoPartialProvider,
)
from py_portfolio_index.operators import (
    generate_composite_order_plan,
    OrderElement,
    OrderType,
)
from py_portfolio_index.enums import ProviderType
from py_portfolio_index.constants import Logger
from logging import StreamHandler, DEBUG

Logger.addHandler(StreamHandler())
Logger.setLevel(DEBUG)


def test_composite():
    provider1 = LocalDictProvider(
        holdings=[
            RealPortfolioElement(ticker="AAPL", units=0.5, value=Money(value=50)),
            RealPortfolioElement(ticker="UNIL", units=1.0, value=Money(value=1000)),
        ],
        cash=Money(value=800),
    )
    provider2 = LocalDictNoPartialProvider(
        holdings=[
            RealPortfolioElement(ticker="AAPL", units=1.0, value=Money(value=100)),
        ],
        cash=Money(value=200),
    )

    provider1._price_dict = {"AAPL": 100, "UNIL": 1000, "MSFT": 33}
    provider2._price_dict = provider1._price_dict

    ideal_port = IdealPortfolio(
        holdings=[
            IdealPortfolioElement(ticker="AAPL", weight=0.5),
            IdealPortfolioElement(ticker="MSFT", weight=0.5),
        ]
    )

    composite = CompositePortfolio(
        portfolios=[provider1.get_holdings(), provider2.get_holdings()]
    )

    expected_size = 2000

    composite_order_plan = generate_composite_order_plan(
        composite=composite,
        ideal=ideal_port,
        target_size=expected_size,
        purchase_order_maps=PurchaseStrategy.LARGEST_DIFF_FIRST,
    )

    no_partial = composite_order_plan[LocalDictNoPartialProvider.PROVIDER]
    assert no_partial.to_buy == [
        OrderElement(
            ticker="MSFT",
            order_type=OrderType.BUY,
            value=None,
            qty=5,
            price=Money(value=33),
            provider=ProviderType.LOCAL_DICT_NO_PARTIAL,
        ),
        # OrderElement(ticker="AAPL", order_type= OrderType.BUY, value=None, qty=0, provider=Provider.LOCAL_DICT_NO_PARTIAL),
    ]
    partial = composite_order_plan[LocalDictProvider.PROVIDER]
    assert partial.to_buy == [
        OrderElement(
            ticker="AAPL",
            order_type=OrderType.BUY,
            value=Money(value="759.9999999999999644728632120"),
            qty=None,
            price=Money(value=100),
            provider=ProviderType.LOCAL_DICT,
        ),
    ]



from py_portfolio_index.portfolio_providers.common import PriceCache
from random import randint

GENERATED = {}


def generate_price(ticker: str):
    if ticker in GENERATED:
        raise ValueError("Already fetched this price, should be cached")
    GENERATED[ticker] = randint(0, 1000)
    return GENERATED[ticker]


def test_price_cache():
    cache = PriceCache(
        fetcher=lambda tickers, date: {y: generate_price(y) for y in tickers},
        single_fetcher=lambda ticker, date: generate_price(ticker),
    )

    assert cache.get_price("AAPL") == cache.get_prices(["AAPL"])["AAPL"]



import contextlib
import json
import multiprocessing
import os
import psutil
import queue
import requests
import sys
import time
import urllib
import urllib3
import warnings


from py_portfolio_index.constants import CACHE_DIR
from platformdirs import user_cache_dir
from pathlib import Path
from dataclasses import dataclass
from os import remove
from typing import TYPE_CHECKING
import atexit

if TYPE_CHECKING:
    from authlib.integrations.httpx_client import OAuth2Client

__TIME_TIME = time.time
CALLBACK_URL = "https://127.0.0.1:8182"
TOKEN_ENDPOINT = "https://api.schwabapi.com/v1/oauth/token"


@dataclass
class SchwabAuthContext:
    authorization_url: str
    api_key: str
    app_secret: str
    callback_url: str
    token_path: Path
    callback_timeout: float
    output_queue: multiprocessing.Queue
    oauth: "OAuth2Client"
    server_pid: int | None


class RedirectTimeoutError(Exception):
    pass


class RedirectServerExitedError(Exception):
    pass


class TokenMetadata:
    """
    Provides the functionality required to maintain and update our view of the
    token's metadata.
    """

    def __init__(self, token, creation_timestamp, unwrapped_token_write_func):
        """
        :param token: The token to wrap in metadata
        :param creation_timestamp: Timestamp at which this token was initially
                                   created. Notably, this timestamp does not
                                   change when the token is updated.
        :unwrapped_token_write_func: Function that accepts a non-metadata
                                     wrapped token and writes it to disk or
                                     other persistent storage.
        """

        self.creation_timestamp = creation_timestamp

        # The token write function is ultimately stored in the session. When we
        # get a new token we immediately wrap it in a new sesssion. We hold on
        # to the unwrapped token writer function to allow us to inject the
        # appropriate write function.
        self.unwrapped_token_write_func = unwrapped_token_write_func

        # The current token. Updated whenever the wrapped token update function
        # is called.
        self.token = token

    @classmethod
    def from_loaded_token(cls, token, unwrapped_token_write_func):
        """
        Returns a new ``TokenMetadata`` object extracted from the metadata of
        the loaded token object. If the token has a legacy format which contains
        no metadata, assign default values.
        """
        if "creation_timestamp" not in token:
            raise ValueError(
                "WARNING: The token format has changed since this token "
                + "was created. Please delete it and create a new one."
            )

        return TokenMetadata(
            token["token"], token["creation_timestamp"], unwrapped_token_write_func
        )

    def token_age(self):
        """Returns the number of second elapsed since this token was initially
        created."""
        return int(time.time()) - self.creation_timestamp

    def wrapped_token_write_func(self):
        """
        Returns a version of the unwrapped write function which wraps the token
        in metadata and updates our view on the most recent token.
        """

        def wrapped_token_write_func(token, *args, **kwargs):
            # If the write function is going to raise an exception, let it do so
            # here before we update our reference to the current token.
            ret = self.unwrapped_token_write_func(
                self.wrap_token_in_metadata(token), *args, **kwargs
            )

            self.token = token

            return ret

        return wrapped_token_write_func

    def wrap_token_in_metadata(self, token):
        return {
            "creation_timestamp": self.creation_timestamp,
            "token": token,
        }


def __update_token(token_path):
    def update_token(t, *args, **kwargs):
        with open(token_path, "w") as f:
            json.dump(t, f)

    return update_token


# This runs in a separate process and is invisible to coverage
def __run_client_from_login_flow_server(
    q, callback_port, callback_path
):  # pragma: no cover
    """Helper server for intercepting redirects to the callback URL. See
    client_from_login_flow for details."""

    import flask

    app = flask.Flask(__name__)

    @app.route(callback_path)
    def handle_token():
        q.put(flask.request.url)
        return "schwab-py callback received! You may now close this window/tab."

    @app.route("/schwab-py-internal/status")
    def status():
        return "running"

    if callback_port == 443:
        return

    # Wrap this call in some hackery to suppress the flask startup messages
    with open(os.devnull, "w") as devnull:
        import logging

        log = logging.getLogger("werkzeug")
        log.setLevel(logging.ERROR)

        old_stdout = sys.stdout
        sys.stdout = devnull
        app.run(port=callback_port, ssl_context="adhoc")
        sys.stdout = old_stdout


def __fetch_and_register_token_from_redirect(
    oauth,
    redirected_url,
    api_key,
    app_secret,
    token_path,
    asyncio,
):
    from schwab.client import AsyncClient, Client
    from schwab.debug import register_redactions
    from authlib.integrations.httpx_client import AsyncOAuth2Client, OAuth2Client

    token = oauth.fetch_token(
        TOKEN_ENDPOINT,
        authorization_response=redirected_url,
        client_id=api_key,
        auth=(api_key, app_secret),
    )

    # Don't emit token details in debug logs
    register_redactions(token)

    # Set up token writing and perform the initial token write
    update_token = __update_token(token_path)
    metadata_manager = TokenMetadata(token, int(time.time()), update_token)
    update_token = metadata_manager.wrapped_token_write_func()
    update_token(token)

    # The synchronous and asynchronous versions of the OAuth2Client are similar
    # enough that can mostly be used interchangeably. The one currently known
    # exception is the token update function: the synchronous version expects a
    # synchronous one, the asynchronous requires an async one. The
    # oauth_client_update_token variable will contain the appropriate one.
    if asyncio:

        async def oauth_client_update_token(t, *args, **kwargs):
            update_token(t, *args, **kwargs)  # pragma: no cover

        session_class = AsyncOAuth2Client
        client_class = AsyncClient
    else:
        oauth_client_update_token = update_token
        session_class = OAuth2Client
        client_class = Client

    # Return a new session configured to refresh credentials
    return client_class(
        api_key,
        session_class(
            api_key,
            client_secret=app_secret,
            token=token,
            update_token=oauth_client_update_token,
            leeway=300,
        ),
        token_metadata=metadata_manager,
        enforce_enums=True,
    )


def create_login_context(
    api_key: str,
    app_secret: str,
    callback_url: str = CALLBACK_URL,
    callback_timeout: float = 300.0,
):
    from authlib.integrations.httpx_client import OAuth2Client
    from schwab import auth

    token_path = (
        Path(user_cache_dir(CACHE_DIR, ensure_exists=True)) / "schwab_token.json"
    )
    try:
        c = auth.client_from_token_file(token_path, api_key, app_secret=app_secret)
        c.get_account_numbers().raise_for_status()
        return
    except FileNotFoundError:
        pass
    except Exception:
        remove(token_path)
        pass
    if callback_timeout is None:
        callback_timeout = 0
    if callback_timeout < 0:
        raise ValueError("callback_timeout must be positive")

    # Start the server
    parsed = urllib.parse.urlparse(callback_url)

    # TODO: move this to validation on the call
    if parsed.hostname != "127.0.0.1":
        # TODO: document this error
        raise ValueError(
            (
                "Disallowed hostname {}. client_from_login_flow only allows "
                + "callback URLs with hostname 127.0.0.1. See here for "
                + "more information: https://schwab-py.readthedocs.io/en/"
                + "latest/auth.html#callback-url-advisory"
            ).format(parsed.hostname)
        )

    callback_port = parsed.port if parsed.port else 443
    callback_path = parsed.path if parsed.path else "/"

    output_queue: multiprocessing.Queue = multiprocessing.Queue()

    server = multiprocessing.Process(
        target=__run_client_from_login_flow_server,
        args=(output_queue, callback_port, callback_path),
    )
    server.start()

    # ensure we clean this up on exit
    def terminate_server():
        try:
            psutil.Process(server.pid).kill()
        except psutil.NoSuchProcess:
            pass

    atexit.register(terminate_server)
    # Wait until the server successfully starts
    while True:
        # Check if the server is still alive
        if server.exitcode is not None:
            # TODO: document this error
            raise RedirectServerExitedError(
                "Redirect server exited. Are you attempting to use a "
                + "callback URL without a port number specified?"
            )
        # Attempt to send a request to the server
        try:
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore", category=urllib3.exceptions.InsecureRequestWarning
                )

                _ = requests.get(
                    "https://127.0.0.1:{}/schwab-py-internal/status".format(
                        callback_port
                    ),
                    verify=False,
                )
            break
        except requests.exceptions.ConnectionError:
            pass

        time.sleep(0.1)

    oauth = OAuth2Client(api_key, redirect_uri=callback_url)
    authorization_url, state = oauth.create_authorization_url(
        "https://api.schwabapi.com/v1/oauth/authorize"
    )

    return SchwabAuthContext(
        authorization_url=authorization_url,
        api_key=api_key,
        app_secret=app_secret,
        callback_url=callback_url,
        token_path=token_path,
        callback_timeout=callback_timeout,
        oauth=oauth,
        output_queue=output_queue,
        server_pid=server.pid,
    )


def fetch_response(context: SchwabAuthContext):
    # Context manager to kill the server upon completion
    @contextlib.contextmanager
    def callback_server():
        try:
            yield
        finally:
            try:
                psutil.Process(context.server_pid).kill()
            except psutil.NoSuchProcess:
                pass

    with callback_server():
        return _fetch_response(context)


def _fetch_response(context: SchwabAuthContext):
    # Wait for a response
    now = __TIME_TIME()
    timeout_time = now + context.callback_timeout
    received_url = None
    while True:
        now = __TIME_TIME()
        if now >= timeout_time:
            if context.callback_timeout == 0:
                # XXX: We're detecting a test environment here to avoid an
                #      infinite sleep. Surely there must be a better way to do
                #      this...
                if __TIME_TIME != time.time:  # pragma: no cover
                    raise ValueError("endless wait requested")
            else:
                break

        # Attempt to fetch from the queue
        try:
            received_url = context.output_queue.get(
                timeout=min(timeout_time - now, 0.1)
            )
            break
        except queue.Empty:
            pass

    if not received_url:
        raise RedirectTimeoutError(
            "Timed out waiting for a post-authorization callback. You "
            + "can set a longer timeout by passing a value of "
            + "callback_timeout to client_from_login_flow."
        )

    return __fetch_and_register_token_from_redirect(
        context.oauth,
        received_url,
        context.api_key,
        context.app_secret,
        context.token_path,
        False,
    )



from py_portfolio_index.datastores.duckdb_datastore import DuckDBDatastore
from py_portfolio_index.models import (
    RealPortfolioElement,
    Money,
)
from py_portfolio_index.portfolio_providers.local_dict import (
    LocalDictProvider,
    LocalDictNoPartialProvider,
)


def test_datastore():
    db = DuckDBDatastore("test.db")
    db.reset()

    provider1 = LocalDictProvider(
        holdings=[
            RealPortfolioElement(ticker="AAPL", units=0.5, value=Money(value=50)),
            RealPortfolioElement(ticker="UNIL", units=1.0, value=Money(value=1000)),
        ],
        cash=Money(value=800),
    )
    provider2 = LocalDictNoPartialProvider(
        holdings=[
            RealPortfolioElement(ticker="AAPL", units=1.0, value=Money(value=100)),
        ],
        cash=Money(value=200),
    )

    db.persist_holding_data(provider1.get_holdings().holdings, provider1.PROVIDER)
    db.persist_holding_data(provider2.get_holdings().holdings, provider2.PROVIDER)

    results = db.query(
        """
WHERE symbol.ticker = 'AAPL'
SELECT
    symbol.ticker,
    sum(holdings.qty) as total_holding_qty,
    sum(holdings.value) as total_holding_value
order by symbol.ticker asc;"""
    ).fetchall()
    assert results[0] == ("AAPL", 1.5, 150)



from decimal import Decimal
from datetime import date
from typing import Optional, List, Dict, DefaultDict, Any
from py_portfolio_index.models import (
    RealPortfolio,
    RealPortfolioElement,
    Money,
    ProfitModel,
)
from py_portfolio_index.common import divide_into_batches
from py_portfolio_index.portfolio_providers.base_portfolio import (
    BaseProvider,
    ObjectKey,
)
from py_portfolio_index.exceptions import (
    ConfigurationError,
    PriceFetchError,
    OrderError,
)
from py_portfolio_index.enums import ProviderType

from py_portfolio_index.portfolio_providers.helpers.moomoo import (
    DEFAULT_PORT,
    MooMooProxy,
)
from functools import lru_cache
from os import environ
from collections import defaultdict


FRACTIONAL_SLEEP = 60
BATCH_SIZE = 50

CACHE_PATH = "moo_moo_tickers.json"


class MooMooProvider(BaseProvider):
    """Provider for interacting with stocks held in
    MooMoo
    """

    PROVIDER = ProviderType.MOOMOO
    SUPPORTS_BATCH_HISTORY = 0
    SUPPORTS_FRACTIONAL_SHARES = True
    PASSWORD_ENV = "MOOMOO_PASSWORD"
    ACCOUNT_ENV = "MOOMOO_ACCOUNT"
    TRADE_TOKEN_ENV = "MOOMOO_TRADE_TOKEN"
    DEVICE_ID_ENV = "MOOMOO_DEVICE_ID"
    OPEND_ENV = "MOOMOO_OPEND_PATH"

    Proxy = MooMooProxy

    def __init__(
        self,
        proxy: MooMooProxy,
        account: str | None = None,
        password: str | None = None,
        trade_token: str | None = None,
        quote_provider: BaseProvider | None = None,
        _external_auth: bool = False,
    ):
        from moomoo import (
            OpenSecTradeContext,
            OpenQuoteContext,
            SecurityFirm,
            TrdMarket,
        )

        if not account:
            account = environ.get(self.ACCOUNT_ENV, None)
        if not password:
            password = environ.get(self.PASSWORD_ENV, None)
        if not trade_token:
            trade_token = environ.get(self.TRADE_TOKEN_ENV, "ABC")
        self._trade_token = trade_token
        # if not device_id:
        #     device_id = environ.get(self.DEVICE_ID_ENV, None)
        if not (account and password and trade_token) and not _external_auth:
            raise ConfigurationError(
                "Must provide ALL OF account, password, trade_token, and arguments or set environment variables MOOMOO_ACCOUNT, MOOMOO_PASSWORD, MOOMOO_TRADE_TOKEN"
            )
        self.proxy = proxy
        self.proxy.validate(account=account, pwd=password)
        self._trade_provider = OpenSecTradeContext(
            filter_trdmarket=TrdMarket.US,
            host="localhost",
            port=DEFAULT_PORT,
            security_firm=SecurityFirm.FUTUINC,
        )
        self._quote_context = OpenQuoteContext(host="localhost", port=DEFAULT_PORT)
        BaseProvider.__init__(self, quote_provider=quote_provider)
        self._local_latest_price_cache: Dict[str, Decimal | None] = defaultdict(
            lambda: None
        )

    @lru_cache(maxsize=None)
    def _get_instrument_price(
        self, ticker: str, at_day: Optional[date] = None
    ) -> Optional[Decimal]:
        # TODO: determine if there is a bulk API
        from moomoo import RET_OK, SubType

        if at_day:
            raise NotImplementedError
            # historicals = self._provider.get_bars(
            #     tId=webull_id,
            #     interval="d1",
            #     timeStamp=int(
            #         datetime(
            #             day=at_day.day,
            #             month=at_day.month,
            #             year=at_day.year,
            #             tzinfo=UTC,
            #         ).timestamp()
            #     ),
            # )
            # return Decimal(value=list(historicals.itertuples())[0].vwap)
        else:
            ret_sub, err_message = self._quote_context.subscribe(
                ["US." + ticker], [SubType.TICKER], subscribe_push=False
            )
            # Subscribe to the K line type first. After the subscription is successful, moomoo OpenD will continue to receive pushes from the server, False means that there is no need to push to the script temporarily
            if ret_sub == RET_OK:  # Subscription successful
                ret, data = self._quote_context.get_stock_quote(
                    ["US." + ticker]
                )  # Get real-time data of subscription stock quotes
                if ret == RET_OK:
                    return list(data.itertuples())[0]
            raise PriceFetchError(
                [ticker], f"Subscription failed, could not get price: {err_message}"
            )

    def _buy_instrument(
        self,
        symbol: str,
        qty: Optional[float],
        value: Optional[Money] = None,
    ) -> bool:
        from moomoo import RET_OK, TrdSide, OrderType

        ret, data = self._trade_provider.unlock_trade(
            password=self._trade_token
        )  # If you use a live trading account to place an order, you need to unlock the account first. The example here is to place an order on a paper trading account, and unlocking is not necessary.
        if ret == RET_OK:
            pass
        else:
            raise ConfigurationError(f"unlock trade error: {data}")

        ret, data = self._trade_provider.place_order(
            # price is arbitrary for makret
            price=0.0 if not value else value.value,
            qty=qty,
            code="US." + symbol,
            order_type=OrderType.MARKET,
            trd_side=TrdSide.BUY,
        )
        if ret == RET_OK:
            return True
        else:
            raise OrderError(f"place_order error: {data}")

    def buy_instrument(
        self, ticker: str, qty: Decimal, value: Optional[Money] = None
    ) -> bool:
        if qty:
            orders_kwargs_list: List[Dict[str, Money | None | Decimal]] = [
                {
                    "qty": qty,
                    "value": None,
                }
            ]
        elif value:
            # market orders require quantity not price
            # so convert here
            price = self.get_instrument_price(ticker)
            orders_kwargs_list = [{"qty": value / price, "value": None}]
        else:
            raise OrderError("Must provide either qty or value for order")
        for order_kwargs in orders_kwargs_list:
            return self._buy_instrument(ticker, **order_kwargs)  # type: ignore
        return True

    def get_unsettled_instruments(self) -> set[str]:
        """We need to efficiently bypass
        paginating all orders if possible
        so just check the account info for if there
        is any cash held for orders first"""
        from moomoo import RET_OK
        from moomoo import OrderStatus

        ret, data = self._trade_provider.order_list_query(
            status_filter_list=[
                OrderStatus.SUBMITTED,
                OrderStatus.FILLED_PART,
                OrderStatus.WAITING_SUBMIT,
            ]
        )
        if ret == RET_OK:
            pass
        else:
            raise ConfigurationError("Could not get order list")
        # code is of format US.MSFT, for example
        return set(item.code.split(".")[-1] for item in data.itertuples())

    def _get_stock_info(self, ticker: str) -> dict:
        raise NotImplementedError()
        # info = self._provider.get_ticker_info(ticker)
        # matches = self._provider.find_instrument_data(ticker)
        # for match in matches:
        #     if match["symbol"] == ticker:
        #         return {
        #             "name": match["simple_name"],
        #             "exchange": match["exchange"],
        #             "market": match["market"],
        #             "country": match["country"],
        #             "tradable": bool(match["tradable"]),
        #         }
        # return info

    def _get_portfolio(self):
        from moomoo import RET_OK

        ret, data = self._trade_provider.accinfo_query()
        if ret == RET_OK:
            return list(data.itertuples())[0]

        raise ConfigurationError("Could not get portfolio")

    def _get_positions(self):
        from moomoo import RET_OK

        ret, data = self._trade_provider.position_list_query()
        if ret == RET_OK:
            return list(data.itertuples())

        raise ConfigurationError("Could not get positions")

    def get_holdings(self) -> RealPortfolio:
        accounts_data = self._get_cached_value(
            ObjectKey.ACCOUNT, callable=self._get_portfolio
        )
        my_stocks = self._get_cached_value(
            ObjectKey.POSITIONS, callable=self._get_positions
        )

        unsettled = self._get_cached_value(
            ObjectKey.UNSETTLED, callable=self.get_unsettled_instruments
        )

        pre = {}
        symbols = []
        total_value = Decimal(0.0)
        for row in my_stocks:
            local: Dict[str, Any] = {}
            local["units"] = row.qty
            ticker = row.code.split(".")[-1]
            local["ticker"] = ticker
            symbols.append(ticker)
            local["value"] = Decimal(row.market_val)
            local["weight"] = 0
            pre[ticker] = local
            total_value += local["value"]

        final = []
        for s in symbols:
            local = pre[s]
            local["weight"] = local["value"] / total_value
            local["unsettled"] = s in unsettled
            final.append(local)
        out = [RealPortfolioElement(**row) for row in final]
        # THIS IS US SPECIFIC ATM
        cash = Decimal(accounts_data.us_cash)
        return RealPortfolio(
            holdings=out,
            cash=Money(value=cash),
            provider=self,
            profit_and_loss=self.get_profit_or_loss(),
        )

    def _get_instrument_prices(
        self, tickers: List[str], at_day: Optional[date] = None
    ) -> Dict[str, Optional[Decimal]]:
        batches: List[Dict[str, Optional[Decimal]]] = []
        for list_batch in divide_into_batches(tickers, 1):
            # TODO: determine if there is a bulk API
            ticker: str = list_batch[0]
            rval = self._get_instrument_price(ticker, at_day=at_day)
            batches.append({ticker: rval})
        prices: Dict[str, Optional[Decimal]] = {}
        for fbatch in batches:
            prices = {**prices, **fbatch}
        return prices

    def get_per_ticker_profit_or_loss(self) -> Dict[str, ProfitModel]:
        my_stocks = self._get_cached_value(
            ObjectKey.POSITIONS, callable=self._get_positions
        )
        dividends = self._get_cached_value(
            ObjectKey.DIVIDENDS, callable=self._get_dividends
        )
        output = {}
        for x in my_stocks:
            ticker = x.code.split(".")[-1]
            output[ticker] = ProfitModel(
                appreciation=Money(value=Decimal(x.pl_val)), dividends=dividends[ticker]
            )
        return output

    def _get_dividends(self) -> defaultdict[str, Money]:
        final: DefaultDict[str, Money] = defaultdict(lambda: Money(value=0))
        return final

    def get_dividend_history(self) -> Dict[str, Money]:
        return super().get_dividend_history()

    def get_dividend_details(self, start=None):
        # TODO: get endpoint for this
        # https://github.com/MoomooOpen/py-moomoo-api/issues/12
        return []

    def _shutdown(self):
        self._trade_provider.close()
        self._quote_context.close()
        if self.proxy:
            self.proxy.close()



# Basic historical cache implementation
# Instantiation equires a callable interace to get values at date for a list of tickers
# accepts list of stocks + dates to get values for
# returns these from cache if possible, or for those not found
# calls provider to return prices
from collections import defaultdict
from typing import List, Dict
from datetime import date as datetype
from datetime import datetime
from decimal import Decimal
from py_portfolio_index.exceptions import PriceFetchError

# 1 hour
DEFAULT_TIMEOUT = 60 * 60


class PriceCache(object):
    def __init__(
        self, fetcher, single_fetcher=None, timeout: int = DEFAULT_TIMEOUT
    ) -> None:
        self.fetcher = fetcher
        self.single_fetcher = single_fetcher
        self.store: defaultdict[str, dict[str, Decimal | None]] = defaultdict(dict)
        self.instant_refresh_times: dict[str, datetime] = {}
        self.default_timeout: int = timeout

    @staticmethod
    def date_to_label(date: datetype | None) -> str:
        if not date:
            label = "INSTANT"
        else:
            label = date.isoformat()
        return label

    def get_price(self, ticker: str, date: datetype | None = None) -> Decimal | None:
        """If we have an optimized single stock lookup"""
        label = self.date_to_label(date)
        cached: dict[str, Decimal | None] = self.store[label]
        if ticker in cached and label == "INSTANT":
            if (
                datetime.now() - self.instant_refresh_times[ticker]
            ).seconds > self.default_timeout:
                del cached[ticker]
        if ticker in cached:
            return cached[ticker]
        try:
            price = self.single_fetcher(ticker, date)
            cached[ticker] = price
            if label == "INSTANT":
                self.instant_refresh_times[ticker] = datetime.now()
            return price
        except NotImplementedError:
            return self.get_prices([ticker], date)[ticker]
        except Exception as e:
            raise PriceFetchError([ticker], e)

    def get_prices(
        self, tickers: List[str], date: datetype | None = None
    ) -> Dict[str, Decimal | None]:
        # if no date is provided, assume they want the instantaneous price
        label = self.date_to_label(date)
        cached: dict[str, Decimal | None] = self.store[label]
        found = {k: v for k, v in cached.items() if k in tickers}
        if label == "INSTANT":
            for k, v in self.instant_refresh_times.items():
                if k in tickers:
                    if (datetime.now() - v).seconds > self.default_timeout:
                        found.pop(k, None)
        missing = [x for x in tickers if x not in found]
        if missing:
            try:
                prices: dict[str, Decimal | None] = self.fetcher(missing, date)
            except PriceFetchError:
                raise
            except Exception as e:
                raise PriceFetchError(missing, e)
            for ticker, price in prices.items():
                cached[ticker] = price
                found[ticker] = price
                if label == "INSTANT":
                    self.instant_refresh_times[ticker] = datetime.now()
        return found



from enum import IntEnum, Enum


class ProviderType(str, Enum):
    ALPACA = "alpaca"
    ALPACA_PAPER = "alpaca_paper"
    ROBINHOOD = "robinhood"
    LOCAL_DICT = "local_dict"
    LOCAL_DICT_NO_PARTIAL = "local_dict_no_partial"
    DUMMY = "dummy"
    WEBULL = "webull"
    WEBULL_PAPER = "webull_paper"
    MOOMOO = "moomoo"
    SCHWAB = "schwab"


class OrderType(Enum):
    BUY = "BUY"
    SELL = "SELL"


class ObjectKey(Enum):
    POSITIONS = 0
    DIVIDENDS = 1
    UNSETTLED = 2
    ACCOUNT = 3
    OPEN_ORDERS = 4
    MISC = 5
    DIVIDENDS_DETAIL = 6


class ProviderClass(Enum):
    PAPER = [
        ProviderType.ALPACA_PAPER,
        ProviderType.WEBULL_PAPER,
        ProviderType.LOCAL_DICT,
    ]
    REAL = [
        ProviderType.ALPACA,
        ProviderType.WEBULL,
        ProviderType.ROBINHOOD,
        ProviderType.SCHWAB,
    ]


class PurchaseStrategy(IntEnum):
    CHEAPEST_FIRST = 1
    LARGEST_DIFF_FIRST = 2
    PEANUT_BUTTER = 3


class RoundingStrategy(int, Enum):
    CLOSEST = 1
    FLOOR = 2
    CEILING = 3


class Currency(str, Enum):
    USD = "$"
    EURO = "€"
    POUND = "£"



from dataclasses import dataclass
from typing import Optional, Dict, Union, Mapping, List, Callable
from decimal import Decimal
from math import floor, ceil
from collections import defaultdict
from py_portfolio_index.common import print_per
from py_portfolio_index.constants import Logger
from py_portfolio_index.enums import PurchaseStrategy, RoundingStrategy
from py_portfolio_index.portfolio_providers.base_portfolio import BaseProvider
from py_portfolio_index.exceptions import PriceFetchError
from py_portfolio_index.models import (
    Money,
    ProviderType,
    OrderElement,
    OrderPlan,
    OrderType,
    PortfolioProtocol,
    CompositePortfolio,
)
from .models import IdealPortfolio

MIN_ORDER_SIZE = 2
MIN_ORDER_MONEY = Money(value=MIN_ORDER_SIZE)


@dataclass
class ComparisonResult:
    ticker: str
    model: Decimal
    comparison: Decimal
    actual: Money

    @property
    def diff(self):
        return self.model - Decimal(self.comparison)


def compare_portfolios(
    real: PortfolioProtocol,
    ideal: IdealPortfolio,
    buy_order=PurchaseStrategy.LARGEST_DIFF_FIRST,
    target_size: Optional[Union[Decimal, int]] = None,
):
    output: Dict[str, ComparisonResult] = {}
    diff = Decimal(0.0)
    selling = Decimal(0.0)
    buying = Decimal(0.0)
    target_value: Money = (
        Money(value=Decimal(target_size)) if target_size else real.value
    )
    for value in ideal.holdings:
        comparison = real.get_holding(value.ticker)
        if not comparison:
            percentage = Decimal(0.0)
            actual_value = Money.parse("0.0")
        else:
            percentage = Decimal((comparison.value / target_value).value)
            actual_value = comparison.value
        output[value.ticker] = ComparisonResult(
            ticker=value.ticker,
            model=value.weight,
            comparison=percentage,
            actual=actual_value,
        )
        _diff = Decimal(value.weight) - percentage
        diff += abs(_diff)
        if _diff < 0:
            selling += abs(_diff)
        else:
            buying += abs(_diff)

    Logger.info(
        f"Total portfolio % delta {print_per(diff)}. Overweight {print_per(selling)}, underweight {print_per(buying)}"
    )
    if buy_order == PurchaseStrategy.LARGEST_DIFF_FIRST:
        diff_output: Dict[str, ComparisonResult] = {
            k: v for k, v in sorted(output.items(), key=lambda item: -abs(item[1].diff))
        }
    elif buy_order == PurchaseStrategy.CHEAPEST_FIRST:
        diff_output = {
            k: v for k, v in sorted(output.items(), key=lambda item: abs(item[1].diff))
        }
    else:
        raise ValueError("Invalid purchase strategy")
    to_purchase = {}
    to_sell = {}
    for key, diffvalue in diff_output.items():
        if diffvalue.diff == 0:
            continue
        elif diffvalue.diff < 0:
            diff_text = "Overweight"
            to_sell[key] = (
                target_value * diffvalue.comparison - target_value * diffvalue.model
            )
        else:
            diff_text = "Underweight"
            to_purchase[key] = (
                target_value * diffvalue.model - target_value * diffvalue.comparison
            )
            # to_purchase.append(key)
        Logger.info(
            f"{diff_text} {key}, {print_per(diffvalue.model)} target vs {print_per(diffvalue.comparison)} actual. Should be {target_value * diffvalue.model}, is {diffvalue.actual}"
        )
    return to_purchase, to_sell


def round_with_strategy(to_buy_currency, rounding_strategy: RoundingStrategy) -> Money:
    if rounding_strategy == RoundingStrategy.CLOSEST:
        to_buy_units = Money(value=int(round(to_buy_currency, 0)))
    elif rounding_strategy == RoundingStrategy.FLOOR:
        to_buy_units = Money(value=floor(to_buy_currency))
    elif rounding_strategy == RoundingStrategy.CEILING:
        to_buy_units = Money(value=ceil(to_buy_currency))
    else:
        raise ValueError("Invalid Rounding Strategy")
    return to_buy_units


def round_int_with_strategy(
    to_buy_currency, rounding_strategy: RoundingStrategy
) -> int:
    if rounding_strategy == RoundingStrategy.CLOSEST:
        to_buy_units = int(round(to_buy_currency, 0))
    elif rounding_strategy == RoundingStrategy.FLOOR:
        to_buy_units = int(floor(to_buy_currency))
    elif rounding_strategy == RoundingStrategy.CEILING:
        to_buy_units = int(ceil(to_buy_currency))
    else:
        raise ValueError("Invalid Rounding Strategy")
    return to_buy_units


def generate_auto_target_size(
    real: CompositePortfolio,
    ideal: IdealPortfolio,
) -> Money:
    cash = Money(value=0)
    for input in real.portfolios:
        cash += input.cash
    in_portfolio_value = Money(value=0)
    for value in ideal.holdings:
        comparison = real.get_holding(value.ticker)
        if not comparison:
            continue
        else:
            in_portfolio_value += comparison.value
    return in_portfolio_value + cash


def generate_sell_order(
    key: str,
    prices: Dict[str, Decimal | None],
    target_value: Money,
    diffvalue: ComparisonResult,
    provider: ProviderType | None = None,
) -> OrderElement | None:
    if round(diffvalue.diff, 4) == 0.0000:
        return None
    elif diffvalue.diff < 0:
        sell_target: Money = (
            target_value * diffvalue.comparison - target_value * diffvalue.model
        )
        price = prices.get(key)
        if not price:
            return None
        qty = round_int_with_strategy(sell_target / price, RoundingStrategy.FLOOR)
        sell_target = max(sell_target, MIN_ORDER_MONEY)
        return OrderElement(
            ticker=key,
            value=sell_target,
            order_type=OrderType.SELL,
            qty=qty,
            provider=provider,
        )
    return None


def generate_buy_order(
    min_order_value: Money,
    scaling_factor: Money,
    purchase_power: Money,
    buy_order: PurchaseStrategy,
    key: str,
    prices: Dict[str, Decimal | None],
    target_value: Money,
    diffvalue: ComparisonResult,
    provider: ProviderType | None = None,
    fractional_shares: bool = True,
) -> OrderElement | None:
    if purchase_power <= 0:
        Logger.debug("No more money to spend")
        return None
    if round(diffvalue.diff, 4) == 0.0000:
        return None
    elif not diffvalue.diff > 0:
        return None
    diff_text = "Underweight"
    initial_buy_target: Money = Money(
        value=min(
            target_value * diffvalue.model - target_value * diffvalue.comparison,
            purchase_power,
        )
    )
    if buy_order == PurchaseStrategy.PEANUT_BUTTER:
        if initial_buy_target > 0.0:
            max_value: Decimal = max(
                Decimal(float(initial_buy_target.value)) * scaling_factor.decimal,
                Decimal(1.0),
            )
            initial_buy_target = Money(value=max_value)
    initial_buy_target = max(initial_buy_target, min_order_value)
    _price = prices[key]
    if not _price:
        return None
    price = Money(value=_price)

    if not fractional_shares:
        qty = round_int_with_strategy(
            initial_buy_target / price, RoundingStrategy.FLOOR
        )
        if qty == 0:
            return None
        buy_target = None
    else:
        # if we can use fractional, go with the target price only
        qty = None
        buy_target = initial_buy_target

    Logger.debug(
        f"{diff_text} {key}, {print_per(diffvalue.model)} target vs {print_per(diffvalue.comparison)} actual. Should be {target_value * diffvalue.model}, is {diffvalue.actual}"
    )
    return OrderElement(
        ticker=key,
        value=buy_target,
        qty=qty,
        price=price,
        order_type=OrderType.BUY,
        provider=provider,
    )


def gen_diff_and_scaling(
    buy_order: PurchaseStrategy,
    output: Dict[str, ComparisonResult],
    purchase_power: Money,
    target_value: Money,
    currently_held: Money,
) -> tuple[Money, Dict[str, ComparisonResult]]:
    scaling_factor = Money(value=1.0)

    if buy_order == PurchaseStrategy.LARGEST_DIFF_FIRST:
        diff_output: Dict[str, ComparisonResult] = {
            k: v for k, v in sorted(output.items(), key=lambda item: -abs(item[1].diff))
        }
    elif buy_order == PurchaseStrategy.CHEAPEST_FIRST:
        diff_output = {
            k: v for k, v in sorted(output.items(), key=lambda item: abs(item[1].diff))
        }
    elif buy_order == PurchaseStrategy.PEANUT_BUTTER:
        # divide the difference between where we want to be
        # and where we are
        # across all stocks
        scaling_factor = purchase_power / (target_value - currently_held)

        diff_output = {
            k: v for k, v in sorted(output.items(), key=lambda item: abs(item[1].diff))
        }
    else:
        raise ValueError("Invalid purchase strategy")
    return scaling_factor, diff_output


def generate_order_plan(
    real: PortfolioProtocol,
    ideal: IdealPortfolio,
    price_fetcher: Callable,
    buy_order=PurchaseStrategy.LARGEST_DIFF_FIRST,
    target_size: Optional[Money | float | int] = None,
    purchase_power: Optional[Money | float | int] = None,
    min_order_value: Money = MIN_ORDER_MONEY,
    skip_tickers: Optional[set[str]] = None,
    fractional_shares: bool = True,
    provider: ProviderType | None = None,
    existing_orders: List[OrderElement] | None = None,
    skip_invalid: bool = True,
    include_sell_orders: bool = False,
) -> OrderPlan:
    diff = Decimal(0.0)
    selling = Decimal(0.0)
    buying = Decimal(0.0)
    target_value: Money = Money(value=target_size) if target_size else real.value
    output: Dict[str, ComparisonResult] = {}
    safe_purchase_power: Money = Money(value=purchase_power or target_value)
    currently_held = Money(value=0)
    current_orders = existing_orders or []
    current_order_val_map: dict[str, Money] = defaultdict(lambda: Money(value=0))
    for current_order in current_orders:
        current_order_val_map[current_order.ticker] += current_order.inferred_value
    for value in ideal.holdings:
        if skip_tickers and value.ticker in skip_tickers:
            continue
        comparison = real.get_holding(value.ticker)
        if not comparison:
            actual_value = Money.parse("0.0")
        else:
            actual_value = comparison.value

        if value.ticker in current_order_val_map:
            actual_value += current_order_val_map[value.ticker]

        if actual_value.is_zero:
            percentage = Decimal(0.0)
        else:
            percentage = Decimal((actual_value / target_value).value)
            actual_value = actual_value

        # track how much we currently have
        currently_held += actual_value
        output[value.ticker] = ComparisonResult(
            ticker=value.ticker,
            model=value.weight,
            comparison=percentage,
            actual=actual_value,
        )
        _diff = Decimal(value.weight) - percentage
        diff += round(abs(_diff), 4)
        if _diff == 0:
            continue
        elif _diff < 0:
            selling += abs(_diff)
        else:
            buying += abs(_diff)

    Logger.info(
        f"Total portfolio % delta {print_per(diff)}. Overweight {print_per(selling)}, underweight {print_per(buying)}, have {purchase_power}"
    )

    scaling_factor, diff_output = gen_diff_and_scaling(
        buy_order, output, safe_purchase_power, target_value, currently_held
    )
    to_purchase: list[OrderElement] = []
    to_sell: list[OrderElement] = []
    price_missing: set[str] = set()
    try:
        prices = price_fetcher([*diff_output.keys()])
    except PriceFetchError as e:
        for x in e.tickers:
            price_missing.add(x)
        Logger.info(
            f"Was unable to fetch prices for {price_missing} tickers, adding to skipped."
        )
        if not skip_invalid:
            raise e
        return generate_order_plan(
            real=real,
            ideal=ideal,
            price_fetcher=price_fetcher,
            buy_order=buy_order,
            target_size=target_size,
            purchase_power=purchase_power,
            min_order_value=min_order_value,
            skip_tickers=(
                skip_tickers.union(price_missing) if skip_tickers else price_missing
            ),
            fractional_shares=fractional_shares,
            provider=provider,
            existing_orders=current_orders,
        )

    for key, diffvalue in diff_output.items():
        sell_order = generate_sell_order(
            key=key,
            prices=prices,
            target_value=target_value,
            diffvalue=diffvalue,
            provider=provider,
        )
        if sell_order and include_sell_orders:
            to_sell.append(sell_order)

    for key, diffvalue in diff_output.items():
        buy_order = generate_buy_order(
            min_order_value=min_order_value,
            scaling_factor=scaling_factor,
            purchase_power=safe_purchase_power,
            buy_order=buy_order,
            key=key,
            prices=prices,
            target_value=target_value,
            diffvalue=diffvalue,
            provider=provider,
            fractional_shares=fractional_shares,
        )
        if buy_order:
            if buy_order.value:
                safe_purchase_power = safe_purchase_power - buy_order.value
            elif buy_order.qty:
                safe_purchase_power = safe_purchase_power - (
                    prices[key] * buy_order.qty
                )
            Logger.debug(f"{safe_purchase_power} left - order is {buy_order}")
            to_purchase.append(buy_order)
        if safe_purchase_power <= 0:
            break

    return OrderPlan(to_buy=to_purchase, to_sell=to_sell)


def generate_composite_order_plan(
    composite: CompositePortfolio,
    ideal: IdealPortfolio,
    purchase_order_maps: Mapping[ProviderType, PurchaseStrategy] | PurchaseStrategy,
    target_size: Optional[Money | float | int],
    min_order_value: Money = MIN_ORDER_MONEY,
    safety_threshold: Decimal = Decimal(0.95),
    target_order_size: Optional[Money] = None,
    include_sell_orders: bool = False,
) -> Mapping[ProviderType, OrderPlan]:
    provider_to_portfolio_map = {
        x.provider: x for x in composite.portfolios if x.provider
    }
    if target_order_size:
        purchase_power_money = {}
        for portfolio in composite.portfolios:
            if portfolio.provider:
                local_power = min(portfolio.cash, target_order_size)
                purchase_power_money[portfolio.provider.PROVIDER] = local_power
                target_order_size -= local_power
    else:
        purchase_power_money = {
            x.provider.PROVIDER: x.cash for x in composite.portfolios if x.provider
        }
    Logger.debug(f"Purchase power money is {purchase_power_money}")
    providers: List[BaseProvider] = list(provider_to_portfolio_map.keys())  # type: ignore

    if isinstance(purchase_order_maps, PurchaseStrategy):
        purchase_order_maps = {x.PROVIDER: purchase_order_maps for x in providers}
    processed = set()
    # check each of our p
    output: defaultdict[ProviderType, OrderPlan] = defaultdict(
        lambda: OrderPlan(to_buy=[], to_sell=[])
    )
    skip_tickers: set[str] = set()
    for provider in providers:
        skip_tickers = skip_tickers.union(provider.get_unsettled_instruments())

    purchase_order = sorted(
        providers, key=lambda x: (x.SUPPORTS_FRACTIONAL_SHARES, x.cash)
    )
    orders: list[OrderElement] = []
    for provider in purchase_order:
        provider_purchase_power: Money = purchase_power_money.get(
            provider.PROVIDER
        ) or Money(value=0)
        processed.add(provider.PROVIDER)
        port = provider_to_portfolio_map[provider]
        # build the plan across the _entire_ composite portfolio

        # if we don't know how much cash we have, skip
        Logger.info(
            f"Doing provider {provider.PROVIDER} with {provider_purchase_power}"
        )
        if not port.cash or port.cash <= Money(value=0):
            Logger.info("No cash left to purchase")
            continue

        local_max_spend = port.cash * safety_threshold
        local_purchase_power = min(provider_purchase_power, local_max_spend)

        purchase_plan: OrderPlan = generate_order_plan(
            ideal=ideal,
            real=composite,
            buy_order=purchase_order_maps[provider.PROVIDER],
            # rounding_strategy=RoundingStrategy.CLOSEST,
            target_size=target_size,
            purchase_power=local_purchase_power,
            min_order_value=min_order_value,
            skip_tickers=skip_tickers,
            fractional_shares=provider.SUPPORTS_FRACTIONAL_SHARES,
            price_fetcher=provider.get_instrument_prices,
            provider=provider.PROVIDER,
            existing_orders=orders,
            include_sell_orders=include_sell_orders,
        )
        orders += purchase_plan.all_orders
        output[provider.PROVIDER] += purchase_plan
    return output


def purchase_composite_order_plan(
    orders: Mapping[ProviderType, OrderPlan],
    providers: list[BaseProvider],
    include_sell_orders: bool = False,
):
    for provider in providers:
        if provider.PROVIDER in orders:
            provider.purchase_order_plan(
                orders[provider.PROVIDER], include_sell_orders=include_sell_orders
            )
        else:
            Logger.info(f"Provider {provider.PROVIDER} has no orders to execute")
    return True



from typing import Protocol, List, Any
from py_portfolio_index.enums import ObjectKey, ProviderType
from datetime import datetime
from trilogy import Environment, Dialects, Executor
from pathlib import Path
from trilogy.dialect.config import DuckDBConfig
from py_portfolio_index.models import DividendResult, RealPortfolioElement


class ResultProtocol(Protocol):
    values: List[Any]
    columns: List[str]

    def fetchall(self) -> List[Any]: ...

    def keys(self) -> List[str]: ...


class DBApiConnectionWrapper:
    def __init__(self, dbapi):
        self.dbapi = dbapi

    def connect(self):
        return self.dbapi


class BaseDatastore:
    def __init__(self, duckdb_path: str, debug: bool = False):
        self.duckdb_path = duckdb_path
        self.debug = debug
        self.executor: Executor = self.connect()

        if not self.check_initialized():
            self.initialize()

    def connect(self):
        env = Environment(working_path=Path(__file__).parent)
        hooks = []
        if self.debug:
            from trilogy.hooks.query_debugger import DebuggingHook

            hooks.append(DebuggingHook())
        self.executor = Dialects.DUCK_DB.default_executor(
            environment=env, conf=DuckDBConfig(path=self.duckdb_path)
        )
        for _ in self.executor.parse_file(Path(__file__).parent / "entrypoint.preql"):
            pass
        return self.executor

    def reset(self):
        self.drop()
        self.connect()
        self.initialize()

    def drop(self):
        raise NotImplementedError

    def check_initialized(self) -> bool:
        raise NotImplementedError

    def initialize(self):
        raise NotImplementedError

    def query(self, query: str) -> ResultProtocol:
        if isinstance(query, str):
            return self.executor.execute_text(query)[-1]
        return self.executor.execute_query(query)

    def get_watermarks(
        self, object_key: ObjectKey, provider_type: ProviderType | None = None
    ) -> tuple[datetime | None, datetime | None]:
        if object_key == ObjectKey.DIVIDENDS:
            query = "dividend.date"
        else:
            raise NotImplementedError
        base_query = f"""
        SELECT
            min({query}) as start,
            max({query}) as end;
        """
        if provider_type:
            base_query = (
                f"WHERE dividend.provider.name='{provider_type.value}' " + base_query
            )
        results = list(self.query(base_query).fetchall())
        if not results:
            return None, None
        return results[0][0], results[0][1]

    def persist_dividend_data(self, data: list[DividendResult]):
        raise NotImplementedError

    def persist_holding_data(
        self, data: list[RealPortfolioElement], provider: ProviderType
    ):
        raise NotImplementedError



import socket as socket
import subprocess
from py_portfolio_index.exceptions import (
    ConfigurationError,
    ExtraAuthenticationStepException,
)
from py_portfolio_index.models import LoginResponse, LoginResponseStatus
from py_portfolio_index.portfolio_providers.helpers.telnetlib import Telnet

DEFAULT_PORT = 11111


def check_listening(port: int):
    # Create a TCP socket
    address = "localhost"
    s = socket.socket()
    try:
        s.connect((address, port))
        return True
    except socket.error:
        return False
    finally:
        s.close()


def interactive_login(
    opend_path: str, account: str, pwd: str, lang: str = "en"
) -> "MooMooProxy":
    proxy = MooMooProxy(opend_path)
    if check_listening(DEFAULT_PORT):
        return proxy
    try:
        proxy.start_proxy(opend_path, account, pwd, lang)
    except ExtraAuthenticationStepException:
        factor = input("Input factor: ")
        proxy.submit_mfa(factor)
    proxy.validate(account, pwd)
    return proxy


class MooMooProxy:
    def __init__(self, opend_path: str | None = None):
        self.opend_path = opend_path
        self.process: subprocess.Popen | None = None
        self.lang = "en"
        self.mfa_in_progress = False
        self.telnet_port = 22222
        self.telnet_ip = "127.0.0.1"

    def validate(
        self, account: str | None, pwd: str | None, extra_factor: str | None = None
    ) -> bool:
        if check_listening(DEFAULT_PORT):
            return True
        if not self.process and self.opend_path and account and pwd:
            self.start_proxy(self.opend_path, account, pwd, self.lang)
        if self.mfa_in_progress and extra_factor:
            self.submit_mfa(extra_factor)
        return check_listening(DEFAULT_PORT)

    def submit_mfa(self, code: str):
        if not self.mfa_in_progress:
            return
        if not self.process:
            raise ConfigurationError(
                "MFA is in progress, but no connection is available to submit the code."
            )
        self.send_command(f"input_phone_verify_code -code={code}")
        self.mfa_in_progress = False

    def start_proxy(
        self, path: str, login_account: str, login_pwd: str, lang: str = "en"
    ) -> bool:
        self.process = subprocess.Popen(
            [
                path,
                f"-login_account={login_account}",
                f"-login_pwd={login_pwd}",
                f"-lang={lang}",
                f"-telnet_ip={self.telnet_ip}",
                f"-telnet_port={self.telnet_port}",
                "-console=1",
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        results = self.send_command("show_sub_info")

        if "The remaining quota:" not in results:
            raise ExtraAuthenticationStepException(
                response=LoginResponse(status=LoginResponseStatus.MFA_REQUIRED)
            )
        return self.validate(login_account, login_pwd)

    def send_command(self, command: str):
        with Telnet(
            self.telnet_ip, self.telnet_port
        ) as tn:  # Telnet address is: 127.0.0.1, Telnet port is: 22222
            tn.write(command.encode() + b"\r\n")
            reply = b""
            while True:
                msg = tn.read_until(b"\r\n", timeout=0.5)
                reply += msg
                if msg == b"":
                    break
            output = reply.decode("gb2312")
            print(output)
        return output

    def close(self):
        if self.process:
            self.send_command("exit")
            self.process = None

    def connect(self, account: str, pwd: str):
        if not self.opend_path:
            raise ValueError(
                "Proxy must be given an OpenD path to automatically start a proxy; if you do not want to do this, ensure a MooMoo OpenD proxy is already running."
            )
            # run this command in a subprocess

        return self.start_proxy(self.opend_path, account, pwd)



# from py_portfolio_index import SchwabProvider
# from py_portfolio_index.portfolio_providers.helpers.schwab import fetch_response, create_login_context
# from datetime import date


# def test_schwab_async_login():

#     login_context = create_login_context()
#     # you actually need
#     print(login_context.auth_url)
#     fetch_response(login_context)
#     test = SchwabProvider(external_auth=True)

#     assert test.get_instrument_price("AAPL", date(2021, 1, 1)) is not None



from logging import getLogger, Logger as baseLogger

LOGGER_NAME = "py_portfolio_index"

CACHE_DIR = "py_portfolio_index"

AUTO_TARGET_SIZE = -1

Logger: baseLogger = getLogger(LOGGER_NAME)

UNKNOWN_TICKER = "???"



from decimal import Decimal
from datetime import date, datetime
from typing import Optional, List, Dict, DefaultDict, Any
from py_portfolio_index.constants import Logger, CACHE_DIR
from py_portfolio_index.models import (
    RealPortfolio,
    RealPortfolioElement,
    Money,
    ProfitModel,
)
from py_portfolio_index.common import divide_into_batches

from py_portfolio_index.portfolio_providers.base_portfolio import (
    BaseProvider,
    ObjectKey,
)
from py_portfolio_index.exceptions import ConfigurationError, PriceFetchError
from py_portfolio_index.models import DividendResult
from collections import defaultdict
import uuid
from py_portfolio_index.enums import ProviderType
from functools import lru_cache
from os import environ
from pytz import UTC

FRACTIONAL_SLEEP = 60
BATCH_SIZE = 50

DEFAULT_WEBULL_TIMEOUT = 60
CACHE_PATH = "webull_tickers.json"


class WebullProvider(BaseProvider):
    """Provider for interacting with stocks held in
    Webull
    """

    PROVIDER = ProviderType.WEBULL
    SUPPORTS_BATCH_HISTORY = 0
    PASSWORD_ENV = "WEBULL_PASSWORD"
    USERNAME_ENV = "WEBULL_USERNAME"
    TRADE_TOKEN_ENV = "WEBULL_TRADE_TOKEN"
    DEVICE_ID_ENV = "WEBULL_DEVICE_ID"

    def _get_provider(self):
        from webull import webull  # for paper trading, import 'paper_webull'

        return webull

    def __init__(
        self,
        username: str | None = None,
        password: str | None = None,
        trade_token: str | None = None,
        device_id: str | None = None,
        skip_cache: bool = False,
    ):
        if not username:
            username = environ.get(self.USERNAME_ENV, None)
        if not password:
            password = environ.get(self.PASSWORD_ENV, None)
        if not trade_token:
            trade_token = environ.get(self.TRADE_TOKEN_ENV, None)
        if not device_id:
            device_id = environ.get(self.DEVICE_ID_ENV, None)
        if not (username and password and trade_token and device_id):
            raise ConfigurationError(
                "Must provide ALL OF username, password, trade_token, and device_id arguments or set environment variables WEBULL_USERNAME, WEBULL_PASSWORD, WEBULL_TRADE_TOKEN, and WEBULL_DEVICE_ID "
            )
        webull = self._get_provider()
        self._provider = webull()
        self._provider.timeout = 60
        # we must set both of these to have a valid login
        self._provider._did = device_id
        self._provider._headers["did"] = device_id

        self._provider.login(username=username, password=password)
        self._local_instrument_cache: Dict[str, str] = {}
        if not skip_cache:
            self._load_local_instrument_cache()

        token = self._provider.get_trade_token(trade_token)
        if not token:
            raise ConfigurationError("Could not get trade token with provided password")
        account_info: dict = self._provider.get_account()
        if account_info.get("success") is False:
            raise ConfigurationError(f"Authentication is expired: {account_info}")
        BaseProvider.__init__(self)

    def _load_local_instrument_cache(self):
        from platformdirs import user_cache_dir
        from pathlib import Path
        import json

        path = Path(user_cache_dir(CACHE_DIR, ensure_exists=True))
        file = path / CACHE_PATH
        if not file.exists():
            self._local_instrument_cache = {}
            return
        with open(file, "r") as f:
            self._local_instrument_cache = json.load(f)
            # corruption guard
            if not isinstance(self._local_instrument_cache, dict):
                self._local_instrument_cache = {}

    def _save_local_instrument_cache(self):
        from platformdirs import user_cache_dir
        from pathlib import Path
        import json

        path = Path(user_cache_dir(CACHE_DIR, ensure_exists=True))
        file = path / CACHE_PATH
        with open(file, "w") as f:
            json.dump(self._local_instrument_cache, f)

    @lru_cache(maxsize=None)
    def _get_instrument_price(
        self, ticker: str, at_day: Optional[date] = None
    ) -> Optional[Decimal]:
        # TODO: determine if there is a bulk API
        webull_id = self._local_instrument_cache.get(ticker)
        if not webull_id:
            # skip the call
            lookup_ticker = ticker.replace(".", "-")
            webull_id = str(self._provider.get_ticker(lookup_ticker))
            self._local_instrument_cache[ticker] = webull_id
            self._save_local_instrument_cache()
        if at_day:
            historicals = self._provider.get_bars(
                tId=webull_id,
                interval="d1",
                timeStamp=int(
                    datetime(
                        day=at_day.day,
                        month=at_day.month,
                        year=at_day.year,
                        tzinfo=UTC,
                    ).timestamp()
                ),
            )
            return Decimal(value=list(historicals.itertuples())[0].vwap)
        else:
            quotes: dict = self._provider.get_quote(tId=webull_id)
            if not quotes.get("askList"):
                return None
            rval = Decimal(quotes["askList"][0]["price"])
            return rval

    def _buy_instrument(
        self, symbol: str, qty: Optional[float], value: Optional[Money] = None
    ) -> dict:
        from webull import webull
        import requests

        # we should always have this at this point, as we would have had
        # to check price
        rtId: Optional[str] = self._local_instrument_cache.get(symbol)
        if not rtId:
            # webull uses '-' instead of '.'; BRK.B -> BRK-B
            lookup_symbol = symbol.replace(".", "-")
            tId = self._provider.get_ticker(lookup_symbol)
            self._local_instrument_cache[symbol] = tId
            self._save_local_instrument_cache()
        else:
            tId = rtId

        def place_order(
            provider: webull,
            tId=tId,
            price=value,
            action="BUY",
            orderType="LMT",
            enforce="GTC",
            quant=qty,
            outsideRegularTradingHour=True,
            stpPrice=None,
            trial_value=0,
            trial_type="DOLLAR",
        ):
            """
            Place an order - redefined here to

            price: float (LMT / STP LMT Only)
            action: BUY / SELL / SHORT
            ordertype : LMT / MKT / STP / STP LMT / STP TRAIL
            timeinforce:  GTC / DAY / IOC
            outsideRegularTradingHour: True / False
            stpPrice: float (STP / STP LMT Only)
            trial_value: float (STP TRIAL Only)
            trial_type: DOLLAR / PERCENTAGE (STP TRIAL Only)
            """

            headers = provider.build_req_headers(
                include_trade_token=True, include_time=True
            )
            data = {
                "action": action,
                "comboType": "NORMAL",
                "orderType": orderType,
                "outsideRegularTradingHour": outsideRegularTradingHour,
                "quantity": quant if orderType == "MKT" else int(quant),
                "serialId": str(uuid.uuid4()),
                "tickerId": tId,
                "timeInForce": enforce,
            }

            # Market orders do not support extended hours trading.
            if orderType == "MKT":
                data["outsideRegularTradingHour"] = False
            elif orderType == "LMT":
                data["lmtPrice"] = float(price)
            elif orderType == "STP":
                data["auxPrice"] = float(stpPrice)
            elif orderType == "STP LMT":
                data["lmtPrice"] = float(price)
                data["auxPrice"] = float(stpPrice)
            elif orderType == "STP TRAIL":
                data["trailingStopStep"] = float(trial_value)
                data["trailingType"] = str(trial_type)
            response = requests.post(
                provider._urls.place_orders(provider._account_id),
                json=data,
                headers=headers,
                timeout=provider.timeout,
            )
            return response.json()

        return place_order(
            self._provider,
            action="BUY",
            price=value,
            quant=qty,
            orderType="MKT",
            enforce="DAY",
        )

    def buy_instrument(self, ticker: str, qty: Decimal, value: Optional[Money] = None):
        if qty:
            float_qty = float(qty)
            import math

            if float_qty > 1:
                remainder_part, int_part = math.modf(float_qty)

                orders = [int(int_part), round(remainder_part, 4)]
            else:
                orders = [float_qty]
            orders_kwargs_list: List[Dict[str, float | Money | None]] = [
                {"qty": order, "value": None} for order in orders
            ]
        else:
            orders_kwargs_list = [{"qty": None, "value": value}]
        for order_kwargs in orders_kwargs_list:
            output = self._buy_instrument(ticker, **order_kwargs)  # type: ignore
            msg = output.get("msg")
            if not output.get("success"):
                if msg:
                    Logger.error(msg)
                    if "Your session has expired" in str(msg):
                        raise ConfigurationError(msg)
                    raise ValueError(msg)
                Logger.error(output)
                if "Your session has expired" in str(output):
                    raise ConfigurationError(output)
                raise ValueError(output)
        return True

    def get_unsettled_instruments(self) -> set[str]:
        orders = self._provider.get_current_orders()
        return set(item["ticker"]["symbol"] for item in orders)

    def _get_stock_info(self, ticker: str) -> dict:
        info = self._provider.get_ticker_info(ticker)
        return info

    def get_portfolio(self) -> dict:
        try:
            return self._provider.get_portfolio()
        except Exception as e:
            raise ConfigurationError(
                f"Could not fetch portfolio on {str(e)}; assuming session expired"
            )

    def get_positions(self) -> list[dict]:
        try:
            return self._provider.get_positions()
        except Exception as e:
            raise ConfigurationError(
                f"Could not fetch positions on {str(e)}; assuming session expired"
            )

    def get_holdings(self) -> RealPortfolio:
        accounts_data = self._get_cached_value(
            ObjectKey.ACCOUNT, callable=self.get_portfolio
        )
        my_stocks = self._get_cached_value(
            ObjectKey.POSITIONS, callable=self.get_positions
        )
        unsettled = self._get_cached_value(
            ObjectKey.UNSETTLED, callable=self.get_unsettled_instruments
        )

        pre = {}
        symbols = []
        for row in my_stocks:
            local: Dict[str, Any] = {}
            local["units"] = row["position"]
            ticker = row["ticker"]["symbol"]
            local["ticker"] = ticker
            symbols.append(ticker)
            local["value"] = 0
            local["weight"] = 0
            pre[ticker] = local
        prices = self._price_cache.get_prices(symbols)
        total_value = Decimal(0.0)
        for s in symbols:
            price = prices[s]
            if not price:
                continue
            total_value += price * Decimal(pre[s]["units"])
        final = []
        pl_info = self.get_per_ticker_profit_or_loss()
        for s in symbols:
            local = pre[s]
            value = Decimal(prices[s] or 0) * Decimal(pre[s]["units"])
            local["value"] = Money(value=value)
            if value == 0.0000:
                local["weight"] = 0.0000
            else:
                local["weight"] = value / total_value
            local["unsettled"] = s in unsettled
            local["appreciation"] = pl_info[s].appreciation
            local["dividends"] = pl_info[s].dividends
            final.append(local)
        out = [RealPortfolioElement(**row) for row in final]
        cash = Decimal(accounts_data["cashBalance"])
        return RealPortfolio(holdings=out, cash=Money(value=cash), provider=self)

    def _get_instrument_prices(
        self, tickers: List[str], at_day: Optional[date] = None
    ) -> Dict[str, Optional[Decimal]]:
        batches: List[Dict[str, Optional[Decimal]]] = []
        if at_day:
            batch_size = 1
        else:
            batch_size = 100
        for list_batch in divide_into_batches(tickers, batch_size):
            # TODO: determine if there is a bulk API
            wb_ids: Dict[str, str] = {}
            new_ids = False
            for ticker in list_batch:
                lookup_ticker = ticker.replace(".", "-")

                webull_id = self._local_instrument_cache.get(ticker)
                if not webull_id:
                    # skip the call
                    try:
                        webull_id = str(self._provider.get_ticker(lookup_ticker))
                    except Exception as e:
                        raise PriceFetchError([ticker], e)
                    self._local_instrument_cache[ticker] = webull_id
                    new_ids = True

                wb_ids[webull_id] = ticker
            if new_ids:
                self._save_local_instrument_cache()
            if at_day:
                output: dict[str, Decimal | None] = {}
                for _, ticker in wb_ids.items():
                    historicals = self._provider.get_bars(
                        stock=ticker,
                        interval="d1",
                        timeStamp=int(
                            datetime(
                                day=at_day.day,
                                month=at_day.month,
                                year=at_day.year,
                                tzinfo=UTC,
                            ).timestamp()
                        ),
                    )
                    base = list(historicals.itertuples())[0]
                    if base:
                        output[ticker] = Decimal(value=base[0].vwap)
                    else:
                        output[ticker] = None
                batches.append(output)
            else:
                from concurrent.futures import ThreadPoolExecutor, as_completed

                final: Dict[str, Decimal | None] = {}
                with ThreadPoolExecutor(max_workers=10) as executor:
                    futures = {
                        executor.submit(self._provider.get_quote, None, wbid)
                        for wbid in wb_ids
                    }
                    for future in as_completed(futures):
                        future_output = future.result()
                        ticker = wb_ids[str(future_output["tickerId"])]
                        if "askList" in future_output:
                            value = future_output["askList"][0]["price"]
                            final[ticker] = Decimal(value=value)
                        else:
                            final[ticker] = None
                batches.append(final)
        prices: Dict[str, Optional[Decimal]] = {}
        for fbatch in batches:
            prices = {**prices, **fbatch}
        return prices

    def get_per_ticker_profit_or_loss(self) -> Dict[str, ProfitModel]:
        my_stocks = self._get_cached_value(
            ObjectKey.POSITIONS, callable=self._provider.get_positions
        )
        dividends = self._get_dividends()
        base = {
            x["ticker"]["symbol"]: ProfitModel(
                appreciation=Money(value=Decimal(x["unrealizedProfitLoss"])),
                dividends=dividends[x["ticker"]["symbol"]],
            )
            for x in my_stocks
        }
        for k, v in dividends.items():
            if k not in base:
                base[k] = ProfitModel(appreciation=Money(value=0), dividends=v)
        return base

    def _get_dividends(self) -> defaultdict[str, Money]:
        dividends: dict = self._get_cached_value(
            ObjectKey.DIVIDENDS, callable=self._provider.get_dividends
        )
        dlist = dividends.get("dividendList", [])
        base = []
        for item in dlist:
            base.append(
                {
                    "value": Money(value=Decimal(item["dividendAmount"])),
                    "ticker": item["tickerTuple"]["symbol"],
                }
            )
        final: DefaultDict[str, Money] = defaultdict(lambda: Money(value=0))
        for item in base:
            final[item["ticker"]] += item["value"]
        return final

    def get_dividend_details(
        self, start: datetime | None = None
    ) -> list[DividendResult]:
        dividends: dict = self._get_cached_value(
            ObjectKey.DIVIDENDS, callable=self._provider.get_dividends
        )
        dlist = dividends.get("dividendList", [])
        final = []
        for x in dlist:
            id = x["tickerTuple"]["tickerId"]
            paid_date = datetime.strptime(x["payDate"], r"%m/%d/%Y").date()
            if start and paid_date < start.date():
                continue
            epoch = paid_date.day + paid_date.month * 100 + paid_date.year * 10000
            epoch += id * 1000000
            final.append(
                DividendResult(
                    ticker=x["tickerTuple"]["symbol"],
                    amount=Money(value=float(x["dividendAmount"])),
                    date=paid_date,
                    provider=self.PROVIDER,
                    external_id=str(epoch),
                )
            )
        return final


class WebullPaperProvider(WebullProvider):
    PROVIDER = ProviderType.WEBULL_PAPER
    PASSWORD_ENV = "WEBULL_PAPER_PASSWORD"
    USERNAME_ENV = "WEBULL_PAPER_USERNAME"
    TRADE_TOKEN_ENV = "WEBULL_PAPER_TRADE_TOKEN"
    DEVICE_ID_ENV = "WEBULL_PAPER_DEVICE_ID"

    def _get_provider(self):
        from webull import paper_webull

        return paper_webull



from py_portfolio_index.datastores.base_datastore import BaseDatastore
from py_portfolio_index.models import DividendResult, RealPortfolioElement
from py_portfolio_index.enums import ProviderType
from py_portfolio_index.constants import UNKNOWN_TICKER
import hashlib


def get_integer_id(value):
    """
    Generate a consistent integer ID from a hash of the given value.

    :param value: The input value (e.g., string, number) to hash
    :return: An integer ID
    """
    # Ensure the value is a string before hashing
    value_str = str(value).encode("utf-8")
    hash_object = hashlib.sha256(value_str)  # Use SHA-256 hash function
    hash_int = int(hash_object.hexdigest(), 16)  # Convert hex digest to an integer
    return hash_int % (2**31)


def map_provider(ptype: ProviderType):
    return {
        ProviderType.LOCAL_DICT_NO_PARTIAL: -2,
        ProviderType.LOCAL_DICT: -1,
        ProviderType.ROBINHOOD: 1,
        ProviderType.ALPACA: 2,
        ProviderType.WEBULL: 3,
        ProviderType.MOOMOO: 4,
        ProviderType.SCHWAB: 5,
    }[ptype]


class DuckDBDatastore(BaseDatastore):
    EXPECTED_TABLES = ["providers", "dividends", "symbols", "ticker_holdings"]

    def __init__(self, db_path: str, debug: bool = False):
        super().__init__(duckdb_path=db_path, debug=debug)

    def check_initialized(self) -> bool:
        assert self.executor is not None
        results = self.executor.execute_raw_sql("SHOW ALL TABLES;").fetchall()
        tables = [
            # table_name index is 2; ex: [('data', 'main', 'providers', ['id', 'name'], ['INTEGER', 'VARCHAR'], False)]
            row[2]
            for row in results
        ]
        return set(tables) == set(self.EXPECTED_TABLES)

    def drop(self):
        for x in self.EXPECTED_TABLES:
            self.executor.execute_raw_sql(f"DROP TABLE {x} CASCADE")
        self.executor.connection.commit()

    def intialize_tickers(self, commit: bool = True):
        from py_portfolio_index.bin import STOCK_INFO

        self.executor.execute_raw_sql(
            """
        CREATE OR REPLACE TABLE symbols (
            id INTEGER PRIMARY KEY,
            ticker VARCHAR,
            name VARCHAR,
            sector VARCHAR,
            industry VARCHAR,
            state VARCHAR,
            city VARCHAR,
            country VARCHAR
        );
        """
        )

        final = [
            {
                "id": idx,
                "ticker": v.ticker,
                "name": v.name,
                "sector": v.sector,
                "industry": v.industry,
                "state": v.state if (v.state and v.state != "") else v.location,
                "city": v.location,
                "country": v.country,
            }
            for idx, v in enumerate(STOCK_INFO.values())
        ]
        final += [
            {
                "id": -1,
                "ticker": UNKNOWN_TICKER,
                "name": "Unknown",
                "sector": "Unknown",
                "industry": "Unknown",
                "state": "Unknown",
                "city": "Unknown",
                "country": "Unknown",
            }
        ]
        for x in final:
            self.executor.execute_raw_sql(
                "INSERT INTO symbols VALUES (:id, :ticker, :name, :sector, :industry, :state, :city, :country);",
                x,
            )
        if commit:
            self.executor.connection.commit()

    def initialize(self):

        self.executor.execute_raw_sql(
            """
        CREATE OR REPLACE TABLE providers (
            id INTEGER PRIMARY KEY,
            name VARCHAR
        );
        """
        )
        self.executor.execute_raw_sql("INSERT INTO providers VALUES (1, 'Robinhood')")
        self.executor.execute_raw_sql("INSERT INTO providers VALUES (2, 'Alpaca')")
        self.executor.execute_raw_sql("INSERT INTO providers VALUES (3, 'Webull')")
        self.executor.execute_raw_sql("INSERT INTO providers VALUES (4, 'Moomoo')")
        self.executor.execute_raw_sql("INSERT INTO providers VALUES (5, 'Schwab')")

        self.executor.execute_raw_sql(
            """
        CREATE OR REPLACE TABLE dividends (
            id INTEGER PRIMARY KEY,
            provider INTEGER,
            symbol INTEGER,
            dividend FLOAT,
            dividend_date DATE,
            dividend_external_id VARCHAR,
        );
        """
        )

        self.executor.execute_raw_sql(
            """
        CREATE OR REPLACE TABLE ticker_holdings (
            symbol integer,
            provider integer,
            qty float,
            cost_basis float,
            value float,
            PRIMARY KEY (symbol, provider)
        );
        """
        )
        self.intialize_tickers(commit=False)
        self.executor.connection.commit()

    def persist_dividend_data(self, data: list[DividendResult]):
        from py_portfolio_index.bin import STOCK_INFO

        for x in data:
            # calculate ID as date epoch + symbol + provider for dividends
            self.executor.execute_raw_sql(
                """INSERT INTO dividends 
                                          SELECT 
                                          
                                            FLOOR(epoch(:dividend_date) / 86400) *10000+symbols.id*10+:provider,
                                            :provider,
                                            symbols.id,
                                            :dividend,
                                            :dividend_date,
                                            :external_id
                                          from symbols
                                          where symbols.ticker = :ticker
                ON CONFLICT DO NOTHING;
                                          """,
                {
                    "ticker": x.ticker if x.ticker in STOCK_INFO else UNKNOWN_TICKER,
                    "provider": map_provider(x.provider),
                    "dividend": x.amount.value,
                    "dividend_date": x.date,
                    "external_id": x.external_id,
                },
            )
        self.executor.connection.commit()

    def persist_holding_data(
        self, data: list[RealPortfolioElement], provider: ProviderType
    ):
        from py_portfolio_index.bin import STOCK_INFO

        for x in data:
            self.executor.execute_raw_sql(
                """INSERT INTO ticker_holdings
                                          SELECT 
                                            symbols.id,
                                            :provider,
                                            :qty,
                                            :cost_basis,
                                            :value
                                          from symbols
                                          where symbols.ticker = :ticker
                ON CONFLICT DO UPDATE SET qty = EXCLUDED.qty, cost_basis = EXCLUDED.cost_basis, value = EXCLUDED.value;
                                          """,
                {
                    "ticker": x.ticker if x.ticker in STOCK_INFO else UNKNOWN_TICKER,
                    "provider": map_provider(provider),
                    "qty": x.units,
                    "cost_basis": x.value.decimal - x.appreciation.decimal,
                    "value": x.value.value,
                },
            )
        self.executor.connection.commit()

    def close(self):
        self.executor.connection.close()
